from typing import Optional


class BadSource(Exception):
    def __init__(self, message: Optional[str] = None):
        if message is not None:
            self._message = message
        else:
            self._message = (
                "One or more critical problems with data source, see logs for details"
            )

    def __str__(self):
        return self._message



from typing import Generator, Tuple, Optional, Callable

import h5py
import numpy as np
from pint import UndefinedUnitError

from .application_logger import get_logger
from .convert_units import get_to_nanoseconds_conversion_method
from .source_error import BadSource
from .convert_units import iso8601_to_ns_since_epoch


def _get_time_offset_in_ns(time_dataset: h5py.Group) -> int:
    """
    Gives an offset which, when added to data times, results in time relative to unix epoch
    """
    try:
        date_string = time_dataset.attrs["start"]
    except KeyError:
        # If no "start" attribute then times are already relative to unix epoch according to NeXus standard
        return 0
    return iso8601_to_ns_since_epoch(date_string)


class LogDataSource:
    def __init__(self, group: h5py.Group):
        """
        Load data from NXlog in NeXus file
        """
        self._group = group
        self._logger = get_logger()

        if self._has_missing_fields():
            raise BadSource()
        try:
            self._convert_time = self._get_time_unit_converter()
        except UndefinedUnitError:
            self._logger.error(
                f"Unable to publish data from NXlog at {self._group.name} due to unrecognised "
                f"or missing units for time field"
            )
            raise BadSource()

        try:
            self._value_index_reached = -1
            self._value_dataset = self._group["value"]

            self._time_index_reached = -1
            self._time_dataset = self._group["time"]

            self._time_offset_ns = _get_time_offset_in_ns(self._time_dataset)
            self._data_are_chunked = True
            try:
                self._value_chunk_iter = self._value_dataset.iter_chunks()
                self._time_chunk_iter = self._time_dataset.iter_chunks()
            except TypeError:
                self._data_are_chunked = False
                return

            self._current_value_slice = next(self._value_chunk_iter)
            self._value_buffer = self._value_dataset[self._current_value_slice]

            self._current_time_slice = next(self._time_chunk_iter)
            self._time_buffer = self._time_dataset[self._current_time_slice]
        except (StopIteration, ValueError):
            self._logger.warn(
                f"Unable to publish data from NXlog at {self._group.name} due to empty value or time field"
            )
            raise BadSource()

    def get_data(self) -> Generator[Tuple[Optional[np.ndarray], int], None, None]:
        """
        Returns None instead of data when there are no more data
        """
        while True:
            if self._data_are_chunked:
                self._value_index_reached += 1
                if self._value_index_reached == self._current_value_slice[0].stop:
                    self._value_index_reached = -1
                    # read next chunk
                    try:
                        self._current_value_slice = next(self._value_chunk_iter)
                    except StopIteration:
                        break
                    self._value_buffer = self._value_dataset[self._current_value_slice]

                self._time_index_reached += 1
                if self._time_index_reached == self._current_time_slice[0].stop:
                    self._time_index_reached = -1
                    # read next chunk
                    try:
                        self._current_time_slice = next(self._time_chunk_iter)
                    except StopIteration:
                        break
                    self._time_buffer = self._time_dataset[self._current_time_slice]

                yield self._value_buffer[self._value_index_reached], self._convert_time(
                    self._time_buffer[self._value_index_reached]
                ) + self._time_offset_ns
            else:
                self._value_index_reached += 1
                if (
                    self._value_index_reached == self._value_dataset.size
                    or self._value_index_reached == self._time_dataset.size
                ):
                    break
                yield self._value_dataset[
                    self._value_index_reached
                ], self._convert_time(
                    self._time_dataset[self._value_index_reached]
                ) + self._time_offset_ns

        yield None, 0

    def _has_missing_fields(self) -> bool:
        missing_field = False
        required_fields = (
            "time",
            "value",
        )
        for field in required_fields:
            if field not in self._group:
                self._logger.error(
                    f"Unable to publish data from NXlog at {self._group.name} due to missing {field} field"
                )
                missing_field = True
        return missing_field

    def _get_time_unit_converter(self) -> Callable:
        try:
            units = self._group["time"].attrs["units"]
        except AttributeError:
            raise UndefinedUnitError
        return get_to_nanoseconds_conversion_method(units)

    @property
    def name(self):
        group_name = self._group.name.split("/")[-1]
        # if group name is "value_log" then use parent group name instead
        # all sample env logs in ISIS files have name value_log...
        if group_name != "value_log":
            return group_name
        return self._group.name.split("/")[-2]



import logging
import configargparse


def parse_args():
    parser = configargparse.ArgumentParser(description="NeXus Streamer")
    parser.add_argument(
        "--graylog-logger-address",
        required=False,
        help="<host:port> Log to Graylog",
        type=str,
        env_var="GRAYLOG_LOGGER_ADDRESS",
    )
    parser.add_argument(
        "--log-file", required=False, help="Log filename", type=str, env_var="LOG_FILE"
    )
    parser.add_argument(
        "-c",
        "--config-file",
        required=False,
        is_config_file=True,
        help="Read configuration from an ini file",
        env_var="CONFIG_FILE",
    )
    log_choice_to_enum = {
        "Trace": logging.DEBUG,
        "Debug": logging.DEBUG,
        "Warning": logging.WARNING,
        "Error": logging.ERROR,
        "Critical": logging.CRITICAL,
    }
    parser.add_argument(
        "-v",
        "--verbosity",
        required=False,
        help="Set logging level",
        choices=log_choice_to_enum.keys(),
        default="Error",
        env_var="VERBOSITY",
    )
    parser.add_argument(
        "-f",
        "--filename",
        required=True,
        help="NeXus file to stream data from",
        env_var="FILENAME",
    )
    parser.add_argument(
        "--json-description",
        required=False,
        help="If provided use this JSON template instead of generating one from the NeXus file",
        env_var="JSON_FILENAME",
    )
    parser.add_argument(
        "-b",
        "--broker",
        required=True,
        help="<host[:port]> Kafka broker to forward data into",
        type=str,
        env_var="BROKER",
    )
    parser.add_argument(
        "-i",
        "--instrument",
        required=True,
        help="Used as prefix for topic names",
        type=str,
        env_var="INSTRUMENT",
    )
    parser.add_argument(
        "-s",
        "--slow",
        action="store_true",
        help="Stream data into Kafka at approx realistic rate (uses timestamps from file)",
        env_var="SLOW",
    )
    parser.add_argument(
        "-z",
        "--single-run",
        action="store_true",
        help="Publish only a single run (otherwise repeats until interrupted)",
        env_var="SINGLE_RUN",
    )
    parser.add_argument(
        "--isis-file",
        action="store_true",
        help="Include ISIS-specific data in event data messages and "
        "detector-spectrum map if found in file",
        env_var="ISIS_FILE",
    )
    parser.add_argument(
        "-e",
        "--fake-events-per-pulse",
        help="Generates this number of fake events per pulse per"
        "event data group instead of publishing real data from file",
        type=int,
        env_var="FAKE_EVENTS",
    )
    parser.add_argument(
        "-d",
        "--det-spec-map",
        help="Full path of a detector-spectrum map file which may be required for files from ISIS",
        env_var="DET_SPEC_MAP",
    )

    optargs = parser.parse_args()
    optargs.verbosity = log_choice_to_enum[optargs.verbosity]
    return optargs



import confluent_kafka
from threading import Thread
from .application_logger import setup_logger
from typing import Optional


class KafkaProducer:
    def __init__(self, configs: dict):
        self._producer = confluent_kafka.Producer(configs)
        self._cancelled = False
        self._poll_thread = Thread(target=self._poll_loop)
        self._poll_thread.start()
        self.logger = setup_logger()

    def _poll_loop(self):
        while not self._cancelled:
            self._producer.poll(0.5)

    def close(self):
        self._cancelled = True
        self._poll_thread.join()
        max_wait_to_publish_producer_queue = 2  # seconds
        self._producer.flush(max_wait_to_publish_producer_queue)

    def produce(
        self,
        topic: str,
        payload: bytes,
        timestamp_ns: Optional[int] = None,
    ):
        def ack(err, _):
            if err:
                self.logger.error(f"Message failed delivery: {err}")

        try:
            if timestamp_ns is not None:
                self._producer.produce(
                    topic,
                    payload,
                    on_delivery=ack,
                    timestamp=int(timestamp_ns * 0.000001),  # ns to ms
                )
            else:
                self._producer.produce(topic, payload, on_delivery=ack)
            self._producer.poll(0)
        except BufferError:
            self.logger.error("Producer buffer full")



from typing import Union
from .event_data_source import EventDataSource, FakeEventDataSource
from .log_data_source import LogDataSource
from .isis_data_source import IsisDataSource

DataSource = Union[LogDataSource, EventDataSource, FakeEventDataSource, IsisDataSource]



from .parse_commandline_args import parse_args
from .application_logger import setup_logger
from .create_data_sources_from_nexus import (
    create_data_sources_from_nexus_file,
)
import h5py
from .kafka_producer import KafkaProducer
from .source_to_stream import (
    LogSourceToStream,
    EventSourceToStream,
    SourceToStream,
)
from .publish_run_message import (
    publish_run_start_message,
    publish_run_stop_message,
)

from .generate_json_description import nexus_file_to_json_description
from .create_data_sources_from_nexus import get_recorded_run_start_time_ns
from typing import List
import asyncio
from time import time_ns
from .isis_data_source import IsisDataSource
from .read_detector_spectrum_map_file import read_map


async def publish_run(producer: KafkaProducer, run_id: int, args, logger):
    streamers: List[SourceToStream] = []
    try:
        recorded_run_start_time_ns, run_start_ds_path = get_recorded_run_start_time_ns(
            args.filename
        )
        # Time difference between starting to stream with NeXus Streamer and the run start time which was recorded
        # in the NeXus file, used as an offset for all timestamps so that output appears as if the data is being
        # produced live by the beamline
        streamer_start_time = time_ns()
        start_time_delta_ns = streamer_start_time - recorded_run_start_time_ns

        log_data_topic = f"{args.instrument}_sampleEnv"
        event_data_topic = f"{args.instrument}_events"
        if args.json_description:
            with open(args.json_description, "r") as json_file:
                nexus_structure = json_file.read()
                nexus_structure = replace_placeholder_topic_names(
                    nexus_structure, log_data_topic, event_data_topic
                )
        else:
            nexus_structure = nexus_file_to_json_description(
                args.filename,
                event_data_topic,
                log_data_topic,
                streamer_start_time,
                run_start_ds_path,
            )

        map_det_ids = None
        map_spec_nums = None
        if args.det_spec_map is not None:
            map_det_ids, map_spec_nums = read_map(args.det_spec_map)

        job_id = publish_run_start_message(
            args.instrument,
            run_id,
            args.broker,
            nexus_structure,
            producer,
            f"{args.instrument}_runInfo",
            map_det_ids,
            map_spec_nums,
        )

        with h5py.File(args.filename, "r") as nexus_file:
            log_data_sources, event_data_sources = create_data_sources_from_nexus_file(
                nexus_file, args.fake_events_per_pulse
            )

            if not log_data_sources and not event_data_sources:
                logger.critical("No valid data sources found in file, aborting")
                return

            isis_data_source = None
            if args.isis_file:
                isis_data_source = IsisDataSource(nexus_file)

            streamers.extend(
                [
                    LogSourceToStream(
                        source,
                        producer,
                        log_data_topic,
                        start_time_delta_ns,
                        slow_mode=args.slow,
                    )
                    for source in log_data_sources
                ]
            )
            streamers.extend(
                [
                    EventSourceToStream(
                        source,
                        producer,
                        event_data_topic,
                        start_time_delta_ns,
                        slow_mode=args.slow,
                        isis_data_source=isis_data_source,
                    )
                    for source in event_data_sources
                ]
            )

            logger.info(
                f"Publishing log data sources: {[source.name for source in log_data_sources]}"
            )
            logger.info(
                f"Publishing event data sources: {[source.name for source in event_data_sources]}"
            )
            for streamer in streamers:
                streamer.start()

            while not all([streamer.done for streamer in streamers]):
                await asyncio.sleep(1.0)

            logger.info("Reached end of data sources")
        publish_run_stop_message(job_id, producer, f"{args.instrument}_runInfo")

    except KeyboardInterrupt:
        logger.info("%% Aborted by user")
    finally:
        for streamer in streamers:
            streamer.stop()
        producer.close()


def replace_placeholder_topic_names(nexus_structure, log_data_topic, event_data_topic):
    for topic in (
        ("SAMPLE_ENV_TOPIC", log_data_topic),
        ("EVENT_DATA_TOPIC", event_data_topic),
    ):
        nexus_structure = nexus_structure.replace(topic[0], topic[1])
    return nexus_structure


def launch_streamer():
    args = parse_args()
    logger = setup_logger(
        level=args.verbosity,
        log_file_name=args.log_file,
        graylog_logger_address=args.graylog_logger_address,
    )
    logger.info("NeXus Streamer started")

    if args.isis_file and not args.det_spec_map:
        logger.warning(
            "ISIS file was specified but no detector-spectrum map was provided,"
            "events may not be mapped to the correct detector pixel by consumer applications"
        )

    producer_config = {
        "bootstrap.servers": args.broker,
        "message.max.bytes": 200000000,
    }
    kafka_producer = KafkaProducer(producer_config)
    run_number = 0
    if args.single_run:
        asyncio.run(publish_run(kafka_producer, run_number, args, logger))
        logger.info(f"Completed streaming run {run_number}")
    else:
        while True:
            asyncio.run(publish_run(kafka_producer, run_number, args, logger))
            logger.info(f"Streamed run {run_number}")
            run_number += 1


if __name__ == "__main__":
    launch_streamer()



import numpy as np
from typing import Tuple


def read_map(file_path: str) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns (detector ids, spectrum numbers)
    """
    det_spec_map = np.genfromtxt(file_path, skip_header=3, dtype=np.int32)
    return det_spec_map[:, 0], det_spec_map[:, 1]



from time import time_ns
import asyncio
from .data_source import (
    EventDataSource,
    LogDataSource,
    FakeEventDataSource,
    IsisDataSource,
)
from typing import Optional, Any, Union
from .kafka_producer import KafkaProducer
from streaming_data_types.logdata_f142 import serialise_f142
from streaming_data_types.eventdata_ev42 import serialise_ev42
import numpy as np


class LogSourceToStream:
    def __init__(
        self,
        source: LogDataSource,
        producer: KafkaProducer,
        output_topic: str,
        start_time_delta_ns: int,
        interval_s: float = 0.2,
        slow_mode: bool = False,
    ):
        """
        :param source: log data source
        :param producer: Kafka producer to use to publish data
        :param output_topic: Kafka topic to publish data to
        :param start_time_delta_ns: diff between time publishing started and start time of the run in the data source
        :param interval_s: idle time between publishing data to allow other async tasks to run
        """
        self._source_name = source.name
        self._data_source = source
        self._producer = producer
        self._topic = output_topic
        self._interval = interval_s
        self._start_time_delta_ns = start_time_delta_ns
        self._cancelled = False
        self._publish_data: Optional[asyncio.Task[Any]] = None
        self._slow_mode = slow_mode

    def start(self):
        self._cancelled = False
        self._publish_data = asyncio.create_task(self._publish_loop())

    def stop(self):
        if not self._cancelled:
            self._cancelled = True
            if self._publish_data is not None:
                self._publish_data.cancel()

    @property
    def done(self):
        return self._cancelled

    async def _publish_loop(self):
        last_timestamp_ns = 0
        get_data = self._data_source.get_data()
        current_run_time_ns = np.iinfo(np.int64).max
        while not self._cancelled:
            if self._slow_mode:
                current_run_time_ns = time_ns()
            while last_timestamp_ns < current_run_time_ns:
                value, data_timestamp_ns = next(get_data)
                if value is not None:
                    if data_timestamp_ns < 0:
                        continue
                    last_timestamp_ns = data_timestamp_ns + self._start_time_delta_ns
                    payload = serialise_f142(
                        value,
                        self._source_name,
                        last_timestamp_ns,
                    )
                    self._producer.produce(
                        self._topic,
                        payload,
                        last_timestamp_ns,
                    )
                else:
                    self._cancelled = True
                    break
            await asyncio.sleep(self._interval)


class EventSourceToStream:
    def __init__(
        self,
        source: Union[EventDataSource, FakeEventDataSource],
        producer: KafkaProducer,
        output_topic: str,
        start_time_delta_ns: int,
        interval_s: float = 0.2,
        slow_mode: bool = False,
        isis_data_source: Optional[IsisDataSource] = None,
    ):
        """
        :param source: event data source
        :param producer: Kafka producer to use to publish data
        :param output_topic: Kafka topic to publish data to
        :param start_time_delta_ns: diff between time publishing started and start time of the run in the data source
        :param interval_s: idle time between publishing data to allow other async tasks to run
        """
        self._source_name = source.name
        self._data_source = source
        self._producer = producer
        self._topic = output_topic
        self._interval = interval_s
        self._start_time_delta_ns = start_time_delta_ns
        self._cancelled = False
        self._publish_data: Optional[asyncio.Task[Any]] = None
        self._message_id = 0
        self._slow_mode = slow_mode
        self._isis_data_source = isis_data_source

    def start(self):
        self._cancelled = False
        self._publish_data = asyncio.create_task(self._publish_loop())

    def stop(self):
        if not self._cancelled:
            self._cancelled = True
            if self._publish_data is not None:
                self._publish_data.cancel()

    @property
    def done(self):
        return self._cancelled

    async def _publish_loop(self):
        last_timestamp_ns = 0
        get_data = self._data_source.get_data()
        if self._isis_data_source is not None:
            get_isis_data = self._isis_data_source.get_data()
        current_run_time_ns = np.iinfo(np.int64).max
        isis_data = None
        while not self._cancelled:
            if self._slow_mode:
                current_run_time_ns = time_ns()
            while last_timestamp_ns < current_run_time_ns:
                time_of_flight, detector_id, data_timestamp_ns = next(get_data)
                if self._isis_data_source is not None:
                    isis_data = next(get_isis_data)
                if time_of_flight is not None:
                    last_timestamp_ns = data_timestamp_ns + self._start_time_delta_ns
                    payload = serialise_ev42(
                        self._source_name,
                        self._message_id,
                        last_timestamp_ns,
                        time_of_flight,
                        detector_id,
                        isis_specific=isis_data,
                    )
                    self._producer.produce(
                        self._topic,
                        payload,
                        last_timestamp_ns,
                    )
                    self._message_id += 1
                else:
                    self._cancelled = True
                    break
            await asyncio.sleep(self._interval)


SourceToStream = Union[LogSourceToStream, EventSourceToStream]



from typing import Generator, Dict, Optional
import h5py
from .application_logger import get_logger
from logging import Logger
import numpy as np


def _get_log_data(
    logger: Logger, nexus_file: h5py.File, log_dataset_path: str, log_dataset_name: str
) -> Optional[np.ndarray]:
    try:
        log_dataset = nexus_file[log_dataset_path]
        return log_dataset[...]
    except KeyError:
        logger.warning(
            f"No {log_dataset_name} dataset found at {log_dataset_path} in ISIS file"
        )
    return None


class IsisDataSource:
    def __init__(self, nexus_file: h5py.File):
        logger = get_logger()

        self._proton_charge_log = _get_log_data(
            logger,
            nexus_file,
            "raw_data_1/framelog/proton_charge/value",
            "proton_charge",
        )
        self._period_log = _get_log_data(
            logger, nexus_file, "raw_data_1/framelog/period_log/value", "period_log"
        )

    def get_data(
        self,
    ) -> Generator[Dict, None, None]:
        isis_specific_data = {
            "period_number": 1,
            "run_state": 1,  # RUNNING
            "proton_charge": 0,
        }
        frame_number = 0
        while True:
            try:
                if self._proton_charge_log is not None:
                    isis_specific_data["proton_charge"] = self._proton_charge_log[
                        frame_number
                    ]
                else:
                    pass
            except IndexError:
                isis_specific_data["proton_charge"] = 0

            try:
                if self._period_log is not None:
                    isis_specific_data["period_number"] = self._period_log[frame_number]
                else:
                    pass
            except IndexError:
                isis_specific_data["period_number"] = 1

            yield isis_specific_data
            frame_number += 1



import pint
from typing import Callable, Union
from pint.errors import UndefinedUnitError
import numpy as np
from datetime import datetime, timezone
import dateutil.parser

ureg = pint.UnitRegistry()
SECONDS = ureg("seconds")
MILLISECONDS = ureg("milliseconds")
MICROSECONDS = ureg("microseconds")
NANOSECONDS = ureg("nanoseconds")


def iso8601_to_ns_since_epoch(iso8601_timestamp: Union[str, bytes]) -> int:
    try:
        iso8601_timestamp = str(iso8601_timestamp, encoding="utf8")  # type: ignore
    except TypeError:
        pass
    offset_datetime = dateutil.parser.parse(iso8601_timestamp)
    if offset_datetime.tzinfo is None:
        # Assume time is UTC if it was not explicit
        offset_datetime = offset_datetime.replace(tzinfo=timezone.utc)
    ns_since_unix_epoch = int(
        offset_datetime.timestamp() * 1_000_000_000
    )  # s float to ns int
    return ns_since_unix_epoch


def ns_since_epoch_to_iso8601(ns_since_epoch: int) -> str:
    dt = datetime.fromtimestamp(ns_since_epoch * 0.000000001)
    dt = dt.replace(tzinfo=timezone.utc)
    return dt.isoformat()


def seconds_to_nanoseconds(input_value: Union[float, int, np.ndarray]) -> np.ndarray:
    return (np.array(input_value) * 1_000_000_000).astype(int)


def milliseconds_to_nanoseconds(
    input_value: Union[float, int, np.ndarray]
) -> np.ndarray:
    return (np.array(input_value) * 1_000_000).astype(int)


def microseconds_to_nanoseconds(
    input_value: Union[float, int, np.ndarray]
) -> np.ndarray:
    return (np.array(input_value) * 1_000).astype(int)


def nanoseconds_to_nanoseconds(
    input_value: Union[float, int, np.ndarray]
) -> np.ndarray:
    return np.array(input_value).astype(int)


def get_to_nanoseconds_conversion_method(units: Union[str, bytes]) -> Callable:
    try:
        units = str(units, encoding="utf8")  # type: ignore
    except TypeError:
        pass
    input_units = ureg(units)

    if input_units == SECONDS:
        return seconds_to_nanoseconds
    elif input_units == MILLISECONDS:
        return milliseconds_to_nanoseconds
    elif input_units == MICROSECONDS:
        return microseconds_to_nanoseconds
    elif input_units == NANOSECONDS:
        return nanoseconds_to_nanoseconds
    else:
        raise UndefinedUnitError



from streaming_data_types.run_start_pl72 import serialise_pl72, DetectorSpectrumMap
from streaming_data_types.run_stop_6s4t import serialise_6s4t
from uuid import uuid4
from time import time_ns
from .kafka_producer import KafkaProducer
from typing import Optional
import numpy as np


def publish_run_start_message(
    instrument_name: str,
    run_number: int,
    broker: str,
    nexus_structure: str,
    producer: KafkaProducer,
    topic: str,
    map_det_ids: Optional[np.ndarray],
    map_spec_nums: Optional[np.ndarray],
) -> str:
    filename = f"FromNeXusStreamer_{run_number}.nxs"
    job_id = str(uuid4())
    start_time_ns = time_ns()
    start_time_ms = int(start_time_ns * 0.000001)
    det_spec_map = None
    if map_spec_nums is not None:
        det_spec_map = DetectorSpectrumMap(
            map_spec_nums, map_det_ids, map_spec_nums.size
        )
    run_start_payload = serialise_pl72(
        job_id,
        filename,
        start_time=start_time_ms,
        run_name=str(run_number),
        nexus_structure=nexus_structure,
        instrument_name=instrument_name,
        broker=broker,
        detector_spectrum_map=det_spec_map,
    )
    producer.produce(topic, run_start_payload, start_time_ns)
    return job_id


def publish_run_stop_message(
    job_id: str,
    producer: KafkaProducer,
    topic: str,
) -> str:
    """
    job_id must match the one used in the corresponding run start message
    """
    stop_time_ns = time_ns()
    stop_time_ms = int(stop_time_ns * 0.000001)
    run_stop_payload = serialise_6s4t(job_id, stop_time=stop_time_ms)
    producer.produce(topic, run_stop_payload, stop_time_ns)
    return job_id



import logging
import graypy
from typing import Optional

logger_name = "nexus-streamer"


def setup_logger(
    level: int = logging.DEBUG,
    log_file_name: Optional[str] = None,
    graylog_logger_address: Optional[str] = None,
) -> logging.Logger:
    if log_file_name is not None:
        logging.basicConfig(filename=log_file_name, format="%(asctime)s %(message)s")
    else:
        logging.basicConfig()
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    if graylog_logger_address is not None:
        host, port = graylog_logger_address.split(":")
        handler = graypy.GELFTCPHandler(host, int(port))
        logger.addHandler(handler)
    return logger


def get_logger():
    return logging.getLogger(logger_name)



import h5py
from typing import Tuple, Optional, Generator, Callable, Union
import numpy as np
from .application_logger import get_logger
from .convert_units import (
    get_to_nanoseconds_conversion_method,
    iso8601_to_ns_since_epoch,
)
from pint.errors import UndefinedUnitError
from .source_error import BadSource


class _ChunkDataLoader:
    def __init__(self, dataset: h5py.Dataset):
        self._dataset = dataset
        self._chunk_iterator = self._dataset.iter_chunks()
        next_slice = next(self._chunk_iterator)
        self._current_chunk = self._dataset[next_slice]
        self._start_index: int = 0

    def get_data_for_pulse(
        self, pulse_start_event: int, pulse_end_event: int
    ) -> np.ndarray:
        start_index = int(pulse_start_event - self._start_index)
        end_index = int(pulse_end_event - self._start_index)

        data_for_pulse = np.array([], dtype=self._current_chunk.dtype)
        while True:
            # If all the data we need is in the current, cached chunk,
            # then just append and return it
            if end_index < self._current_chunk.size:
                return np.append(
                    data_for_pulse, self._current_chunk[start_index:end_index]
                )
            # else...
            # we need all the data in the current chunk...
            data_for_pulse = np.append(
                data_for_pulse, self._current_chunk[start_index:]
            )
            # and at least some from the next chunk, so load the next chunk and continue
            end_index -= self._current_chunk.size
            start_index = 0
            self._start_index += self._current_chunk.size
            try:
                next_slice = next(self._chunk_iterator)
            except StopIteration:
                return data_for_pulse
            self._current_chunk = self._dataset[next_slice]


class _ContiguousDataLoader:
    def __init__(self, dataset: h5py.Dataset):
        self._dataset = dataset
        max_bytes_willing_to_load_into_memory = 100_000_000  # 100 MB
        if self._dataset.nbytes < max_bytes_willing_to_load_into_memory:
            self._dataset = self._dataset[...]
        elif self._dataset.compression is not None:
            get_logger().warning(
                f"{self._dataset.name} is larger than {max_bytes_willing_to_load_into_memory} bytes,"
                f"contiguous and compressed, it will be very slow to stream if these event data are from many pulses"
            )

    def get_data_for_pulse(
        self, pulse_start_event: int, pulse_end_event: int
    ) -> np.ndarray:
        return self._dataset[pulse_start_event:pulse_end_event]


_DataLoader = Union[_ChunkDataLoader, _ContiguousDataLoader]


def _get_pulse_time_offset_in_ns(pulse_time_dataset: h5py.Group) -> int:
    """
    Gives an offset which, when added to pulse times, results in time relative to unix epoch
    """
    try:
        date_string = pulse_time_dataset.attrs["offset"]
    except KeyError:
        # If no "offset" attribute then times are already relative to unix epoch according to NeXus standard
        return 0
    return iso8601_to_ns_since_epoch(date_string)


class EventDataSource:
    def __init__(self, group: h5py.Group):
        """
        Load data, one pulse at a time from NXevent_data in NeXus file
        :raises BadSource if there is a critical problem with the data source
        """
        self._group = group
        self._logger = get_logger()
        self._tof_loader: _DataLoader
        self._id_loader: _DataLoader

        if self._has_missing_fields():
            raise BadSource()
        try:
            self._convert_pulse_time = _get_pulse_time_unit_converter(group)
            self._convert_event_time = self._get_event_time_unit_converter()
        except UndefinedUnitError:
            self._logger.error(
                f"Unable to publish data from NXevent_data at {self._group.name} due to unrecognised "
                f"or missing units for time field"
            )
            raise BadSource()

        self._event_time_zero = self._group["event_time_zero"][...]
        self._event_index = self._group["event_index"][...]

        # There is some variation in the last recorded event_index in files from different institutions
        # for example ISIS files often have what would be the first index of the next pulse at the end.
        # This logic hopefully covers most cases
        if self._event_index[-1] < self._group["event_id"].len():
            self._event_index = np.append(
                self._event_index,
                np.array([self._group["event_id"].len() - 1]).astype(
                    self._event_index.dtype
                ),
            )
        else:
            self._event_index[-1] = self._group["event_id"].len()

        try:
            self._group["event_time_offset"].iter_chunks()
            self._tof_loader = _ChunkDataLoader(self._group["event_time_offset"])
        except TypeError:
            self._tof_loader = _ContiguousDataLoader(self._group["event_time_offset"])

        try:
            self._group["event_id"].iter_chunks()
            self._id_loader = _ChunkDataLoader(self._group["event_id"])
        except TypeError:
            self._id_loader = _ContiguousDataLoader(self._group["event_id"])

        self._pulse_time_offset_ns = _get_pulse_time_offset_in_ns(
            self._group["event_time_zero"]
        )

    def get_data(
        self,
    ) -> Generator[Tuple[Optional[np.ndarray], Optional[np.ndarray], int], None, None]:
        """
        Returns None instead of a data when there is no more data
        """
        # -1 as last index would be start of the next pulse after the end of the run
        for pulse_number in range(self._event_index.size - 1):
            pulse_time = (
                self._convert_pulse_time(self._event_time_zero[pulse_number])
                + self._pulse_time_offset_ns
            )
            start_event = self._event_index[pulse_number]
            end_event = self._event_index[pulse_number + 1]
            yield self._convert_event_time(
                self._tof_loader.get_data_for_pulse(start_event, end_event)
            ), self._id_loader.get_data_for_pulse(start_event, end_event), pulse_time
        yield None, None, 0

    def _has_missing_fields(self) -> bool:
        missing_field = False
        required_fields = (
            "event_time_zero",
            "event_index",
            "event_id",
            "event_time_offset",
        )
        for field in required_fields:
            if field not in self._group:
                self._logger.error(
                    f"Unable to publish data from NXevent_data at {self._group.name} due to missing {field} field"
                )
                missing_field = True
        return missing_field

    def _get_event_time_unit_converter(self) -> Callable:
        try:
            units = self._group["event_time_offset"].attrs["units"]
        except AttributeError:
            raise UndefinedUnitError
        return get_to_nanoseconds_conversion_method(units)

    @property
    def name(self):
        return self._group.name.split("/")[-1]


def _get_pulse_time_unit_converter(group: h5py.Group) -> Callable:
    try:
        units = group["event_time_zero"].attrs["units"]
    except AttributeError:
        raise UndefinedUnitError
    return get_to_nanoseconds_conversion_method(units)


class FakeEventDataSource:
    def __init__(self, group: h5py.Group, events_per_pulse: int):
        logger = get_logger()
        try:
            self._detector_ids = group.parent["detector_number"][...]
        except KeyError:
            logger.error(
                "detector_number dataset not found in parent group of "
                "NXevent_data, this must be present when using "
                "--fake-events-per-pulse"
            )
            raise BadSource()
        self._events_per_pulse = events_per_pulse

        self._event_time_zero = group["event_time_zero"][...]
        self._convert_pulse_time = _get_pulse_time_unit_converter(group)

        self._pulse_time_offset_ns = _get_pulse_time_offset_in_ns(
            group["event_time_zero"]
        )
        self._rng = np.random.default_rng(12345)

        self.name = group.name.split("/")[-1]

    def get_data(
        self,
    ) -> Generator[Tuple[Optional[np.ndarray], Optional[np.ndarray], int], None, None]:
        """
        Returns None instead of a data when there is no more data
        """
        for pulse_number in range(self._event_time_zero.size):
            pulse_time = (
                self._convert_pulse_time(self._event_time_zero[pulse_number])
                + self._pulse_time_offset_ns
            )

            tofs = self._rng.integers(
                low=10000, high=10000000, size=self._events_per_pulse
            )
            detector_num_indices = self._rng.integers(
                low=0, high=self._detector_ids.size, size=self._events_per_pulse
            )
            ids = self._detector_ids[detector_num_indices]

            yield tofs, ids, pulse_time
        yield None, None, 0



import h5py
from typing import Union, Tuple, Dict, List, Optional
from .data_source import EventDataSource, FakeEventDataSource
from .log_data_source import LogDataSource
from .source_error import BadSource
from .application_logger import get_logger
from .convert_units import iso8601_to_ns_since_epoch


def get_attr_as_str(h5_object, attribute_name: str):
    try:
        return h5_object.attrs[attribute_name].decode("utf8")
    except AttributeError:
        return h5_object.attrs[attribute_name]


def find_by_nx_class(
    nx_class_names: Tuple[str, ...], root: Union[h5py.File, h5py.Group]
) -> Dict[str, List[h5py.Group]]:
    groups_with_requested_nx_class: Dict[str, List[h5py.Group]] = {
        class_name: [] for class_name in nx_class_names
    }

    def _match_nx_class(_, h5_object):
        if isinstance(h5_object, h5py.Group):
            try:
                if get_attr_as_str(h5_object, "NX_class") in nx_class_names:
                    groups_with_requested_nx_class[
                        get_attr_as_str(h5_object, "NX_class")
                    ].append(h5_object)
            except KeyError:
                pass

    root.visititems(_match_nx_class)
    return groups_with_requested_nx_class


def create_data_sources_from_nexus_file(
    nexus_file: h5py.File,
    fake_events_per_pulse: Optional[int],
) -> Tuple[List[LogDataSource], List[Union[EventDataSource, FakeEventDataSource]]]:
    nx_log = "NXlog"
    nx_event_data = "NXevent_data"
    groups = find_by_nx_class((nx_log, nx_event_data), nexus_file)

    log_sources = []
    for group in groups[nx_log]:
        try:
            log_sources.append(LogDataSource(group))
        except BadSource:
            # Reason for error is logged in source init
            pass

    event_sources: List[Union[EventDataSource, FakeEventDataSource]] = []
    for group in groups[nx_event_data]:
        try:
            if fake_events_per_pulse is not None:
                event_sources.append(FakeEventDataSource(group, fake_events_per_pulse))
            else:
                event_sources.append(EventDataSource(group))
        except BadSource:
            # Reason for error is logged in source init
            pass

    return log_sources, event_sources


def get_recorded_run_start_time_ns(filename: str) -> Tuple[int, str]:
    logger = get_logger()
    with h5py.File(filename, "r") as nexus_file:
        # start_time dataset in the NXentry is method of recording run start supported by the standard
        nx_entry_class = "NXentry"
        entry_groups = find_by_nx_class((nx_entry_class,), nexus_file)
        if len(entry_groups) > 1:
            logger.warning(
                "More than one NXentry group found."
                "Note: NeXus Streamer will stream data from all groups as if it were one experiment run."
            )
        run_start_time_ns = None
        run_start_dataset_path = ""
        for entry_group in entry_groups[nx_entry_class]:
            try:
                start_time_dataset = entry_group["start_time"]
                found_start_time_ns = iso8601_to_ns_since_epoch(start_time_dataset[...])
                if run_start_time_ns is None or found_start_time_ns < run_start_time_ns:
                    run_start_time_ns = found_start_time_ns
                    run_start_dataset_path = start_time_dataset.name
            except KeyError:
                pass

        if run_start_time_ns is not None:
            return run_start_time_ns, run_start_dataset_path

        logger.error("No start_time dataset found in NXentry to use as run start time")
        raise RuntimeError("Found nothing to use as run start time")
        # TODO find earliest NXlog or NXevent_data timestamp and use that as run start...



import numpy as np
import nexusformat.nexus as nexus
import json
from .convert_units import ns_since_epoch_to_iso8601
from typing import Tuple, Union


# These datasets are always truncated, this avoids JSON descriptions
# of ISIS files in particular from becoming huge
_always_truncate_list = (
    "event_id",
    "event_index",
    "event_time_bins",
    "event_time_offset",
    "event_time_zero",
    "event_frame_number",
    "counts",
)


class NexusToDictConverter:
    """
    Class used to convert nexus format root to python dict
    """

    def __init__(
        self,
        truncate_large_datasets: bool = False,
        large: int = 10,
        event_data_topic: str = "EVENT_DATA_TOPIC",
        log_data_topic: str = "LOG_DATA_TOPIC",
    ):
        """
        :param truncate_large_datasets: if True truncates datasets with any dimension larger than large
        :param large: dimensions larger than this are considered large
        """
        self.truncate_large_datasets = truncate_large_datasets
        self.large = large
        self._event_data_topic = event_data_topic
        self._log_data_topic = log_data_topic

    def convert(self, nexus_root: nexus.NXroot) -> dict:
        """
        Converts the given nexus_root to dict
        """
        return {
            "children": [
                self._root_to_dict(entry) for _, entry in nexus_root.entries.items()
            ]
        }

    def _root_to_dict(self, root: nexus.NXgroup) -> dict:
        try:
            if hasattr(root, "entries"):
                root_dict = self._handle_group(root)
            else:
                root_dict = self._handle_dataset(root)

            root_dict = self._handle_attributes(root, root_dict)
        except nexus.tree.NeXusError:
            root_dict = {}
        return root_dict

    def truncate_if_large(self, size, data, name: str):
        if self.truncate_large_datasets:
            return self.truncate(data, size)
        if name in _always_truncate_list:
            return self.truncate(data, size)
        return data

    def truncate(self, data, size):
        size = list(size)
        for dim_number, dim_size in enumerate(size):
            if dim_size > self.large:
                size[dim_number] = self.large
        data_copy = np.copy(data[...])
        data_copy.resize(size)
        return data_copy

    def _get_data_and_type(self, root, name: str):
        size: Union[int, Tuple[int, ...]] = 1
        data = root.nxdata
        dtype = str(root.dtype)
        if isinstance(data, np.ndarray):
            size = data.shape
            data = self.truncate_if_large(size, data, name)
            if dtype[:2] == "|S":
                data = np.char.decode(data)
            data = data.tolist()
        if dtype[:2] == "|S":
            if not isinstance(data, list):
                data = data.decode("utf-8")
            dtype = "string"
        elif dtype == "float64":
            dtype = "double"
        elif dtype == "float32":
            dtype = "float"
        return data, dtype, size

    @staticmethod
    def _skip_isis_specific_nodes(root):
        if root.nxclass[:2] == "IX":
            return True
        return False

    def _handle_attributes(self, root, root_dict):
        if root.nxclass and root.nxclass != "NXfield" and root.nxclass != "NXgroup":
            root_dict["attributes"] = [{"name": "NX_class", "values": root.nxclass}]
        if root.attrs:
            if "attributes" not in root_dict:
                root_dict["attributes"] = []
            root_dict["attributes"] = []
            for attr_name, attr in root.attrs.items():
                data, dtype, size = self._get_data_and_type(attr, attr_name)
                new_attribute = {"name": attr_name, "values": data}
                if dtype != "object":
                    new_attribute["type"] = dtype
                root_dict["attributes"].append(new_attribute)
        return root_dict

    def _handle_group(self, root):
        root_dict = {"type": "group", "name": root.nxname, "children": []}
        # Add the entries
        entries = root.entries

        if not self._handle_stream(
            root, root_dict
        ) and not self._skip_isis_specific_nodes(root):
            if entries:
                for entry in entries:
                    child_dict = self._root_to_dict(entries[entry])
                    root_dict["children"].append(child_dict)

        return root_dict

    @staticmethod
    def _get_dtype(root) -> str:
        """
        Get the datatype as a string in the format expected by the file writer
        """
        dtype = str(root.dtype)
        if dtype[:2] == "|S":
            dtype = "string"
        elif dtype == "float64":
            dtype = "double"
        elif dtype == "float32":
            dtype = "float"
        return dtype

    def _handle_stream(self, root, root_dict: dict) -> bool:
        """
        Insert information to get this data from Kafka stream if it is of an NXclass supported by the NeXus-Streamer
        :param root:
        :param root_dict:
        :return: true if data will be streamed
        """
        stream_info = {}
        is_stream = False
        if isinstance(root, nexus.NXlog):
            stream_info["writer_module"] = "f142"
            is_stream = True
            if root.nxname == "value_log":
                # For ISIS files the parent of the NXlog has a more useful name
                # which the NeXus-Streamer uses as the source name
                stream_info["source"] = root.nxgroup.nxname
            else:
                stream_info["source"] = root.nxname
            stream_info["topic"] = self._log_data_topic
            try:
                stream_info["dtype"] = self._get_dtype(root.entries["value"])
            except KeyError:
                try:
                    stream_info["dtype"] = self._get_dtype(root.entries["raw_value"])
                except KeyError:
                    is_stream = False
            try:
                stream_info["unit"] = root.attrs["units"]
            except KeyError:
                pass
        elif isinstance(root, nexus.NXevent_data):
            stream_info["writer_module"] = "ev42"
            stream_info["topic"] = self._event_data_topic
            stream_info["source"] = "NeXus-Streamer"
            is_stream = True
        if is_stream:
            root_dict["children"].append({"type": "stream", "stream": stream_info})
        return is_stream

    def _handle_dataset(self, root):
        data, dataset_type, size = self._get_data_and_type(root, root.nxname)
        root_dict = {
            "type": "dataset",
            "name": root.nxname,
            "dataset": {"type": dataset_type},
            "values": data,
        }
        if size != 1:
            root_dict["dataset"]["size"] = size

        return root_dict


def _get_child_node(name: str, tree: dict) -> dict:
    for child in tree["children"]:
        try:
            if child["name"] == name:
                return child
        except KeyError:
            continue
    raise KeyError


def _replace_old_start_time_with_streamer_start_time(
    new_run_start_ns: int, run_start_dataset_path: str, json_tree: dict
):
    new_run_start_time = ns_since_epoch_to_iso8601(new_run_start_ns)
    run_start_dataset_path_list = run_start_dataset_path.split("/")[1:]
    for node_name in run_start_dataset_path_list:
        json_tree = _get_child_node(node_name, json_tree)
    json_tree["values"] = [new_run_start_time]


def nexus_file_to_json_description(
    filename: str,
    event_data_topic: str,
    log_data_topic: str,
    new_run_start_ns: int,
    run_start_dataset_path: str,
) -> str:
    nexus_file = nexus.nxload(filename)
    converter = NexusToDictConverter(
        truncate_large_datasets=False,
        event_data_topic=event_data_topic,
        log_data_topic=log_data_topic,
    )
    tree = converter.convert(nexus_file)
    _replace_old_start_time_with_streamer_start_time(
        new_run_start_ns, run_start_dataset_path, tree
    )

    return json.dumps(tree, indent=2, sort_keys=False)
